package org.w3c.rdfpic.core;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.Writer;
import java.io.File;
import java.net.URL;
import java.util.Vector;

/**
 * A class to store/load schemas.
 *
 * SchemaTxtDrv uses a simple format for storing schema definitions.
 * the following is an example:
 * <p>
 * name = "ThisSchemasname"<br/>
 * URI = "http://thisschema.com/thisschema.def"<br>
 * builtin = "0"<br>
 * add(NAME,EXPLANATION,DEFAULTVAL)<br>
 * <p>
 * Where the parameters of add are all strings. DEFAULTVAL is optional.
 * See the actual schemas used by RdfPic for more examples.
 */
public class SchemaTxtDrv implements SchemaFileDrv
{
    URL       active;
    public void setURL(URL url) {active=url;}

    public Object clone()
    {
	SchemaTxtDrv retval =new SchemaTxtDrv();
	retval.active=active;
	return retval;
    }
    public void write(RdfSchema schema) throws IOException
    {
	if(active ==null) throw new IOException("To use SchemaTxtDrv the URL must be non-null");
	if(!active.getProtocol().equals("file"))
	    throw new IOException("Only able to save schemas to file");
	File outFile = new File(active.getFile());
	Writer writer = new BufferedWriter(new FileWriter(outFile));
	writer.write("name = \""+schema.name+"\"\n");
	writer.write("URI = \""+schema.URI +"\"\n");
	writer.write("builtin = "+schema.builtin +"\n");
	for(int i=0;i<schema.predicates.length;++i) 
	  writer.write("add(\""
		       +(schema.isBuiltin[i] ? "_" : "")
		       +schema.predicates[i]+"\""
		       +(schema.defaults[i]!=null ? ",\""+schema.defaults[i]+"\"" : "")
		       +")\n");
	writer.close();
    }
    public RdfSchema read() throws IOException
    {
	int index;
	boolean error=false;
	if (active==null) throw new IOException("No URL specified");
	char[] whites = {' ','\n','\t'};
	char[] specs  = {'(',')','=',','};
	return (new TxtSchemaParser(
	            new SimpleToken(whites,specs,'"',
		        new BufferedReader(
		            new InputStreamReader(active.openStream()))))).getSchema();
    }
}
/**
 * An exception thrown by the simple parser used for the schema files.
 */
class SyntaxException extends Exception
{
    SyntaxException(String err){super(err);}
}
/**
 * A parser for a simple schema definition language.
 * It closely follow the following:
 * <p>
 *
 *  schema  = [main]<br>
 *  main    = STR | add | namedef | uridef<br>
 *  namedef = 'name' '=' STR<br>
 *  uridef  = 'URI' '=' STR<br>
 *  builtin = 'builtin' '=' INT<br>
 *  add     = 'add' '(' STR ',' STR ',' STR')' | 'add' '(' STR ',' STR ')'<br>
 *  INT     = STR<br>
 * <p>
 * The constructor parses the file immediate, and getSchema() can be used to
 * to retrieve the schema after successful parsing.
 */
class TxtSchemaParser
{
    SimpleToken tokGen;
    String current;
    RdfSchema schema;
    Vector vec;
    Vector defs;
    Vector expl;
    
    TxtSchemaParser(SimpleToken tokGen)
    {
	this.tokGen=tokGen;
	advance();
	schema= new RdfSchema();
	vec = new Vector();
	defs = new Vector();
        expl = new Vector();
        schema.builtin=0;
	try {
	    mainP();
	    if (schema.name==null) schema=null;
	}
	catch (SyntaxException e)
	    {
		schema=null;
	    }
	if(schema!=null)
	    {
		int len = vec.size();
		schema.isBuiltin=new boolean[len];
		for(int i=0;i<len;++i)
		  schema.isBuiltin[i]=((String)vec.elementAt(i)).startsWith("_")
		    && schema.builtin != 0;
		schema.predicates=new String[len];
		for(int i=0;i<len;++i)
		  if (schema.isBuiltin[i])
		    schema.predicates[i]=((String)vec.elementAt(i)).substring(1);
		  else
		    schema.predicates[i]=(String)vec.elementAt(i);
		schema.defaults=new String[len];
		for(int i=0;i<len;++i)
		    schema.defaults[i]=(String)defs.elementAt(i);
		schema.expl=new String[len];
		for(int i=0;i<len;++i)
		    schema.expl[i]=(String)expl.elementAt(i);
	    }
	vec=null;
        defs=null;
    }
    public RdfSchema getSchema()
    {
	return schema;
    }
    private void advance()
    {
	current=tokGen.nextToken();
	//	System.err.print(current+"\n");
    }
    private final void mainP() throws SyntaxException
    {
	while(current!= null)
	    {
		if (!(strP() || addP() || namedefP() || uridefP() || builtinP()))
		    throw new SyntaxException("illegal token");
	    }
    }
    private final boolean isStr()
    {
	if(current.startsWith("\"")&&current.endsWith("\"")) return true;
	return false;
    }
    private final void accept(String str) throws SyntaxException
    {
	if (!str.equals(current)) throw new SyntaxException("not accepted - \"" + str + "\"!=\""+current+"\"");
    }
    private final boolean strP() throws SyntaxException
    {
	if(isStr()) {advance();return true;}
	return false;
    }
    private final String extract() throws SyntaxException
    {
	if (current.length()<2) throw new SyntaxException("short string");
	return current.substring(1,current.length()-1);
    }
    private final void test() throws SyntaxException
    {
	if (current==null) throw new SyntaxException("no more tokens");
    }
    private final boolean addP() throws SyntaxException
    {
	if (current.equals("add"))
	    {
		advance();test();
		accept("(");
		advance();test();
		if(!isStr()) throw new SyntaxException("add: string needed");
		vec.add(extract());
		advance();test();
                accept(",");
                advance();test();
		if(!isStr()) throw new SyntaxException("add: string needed");
		expl.add(extract());
		advance();test();
		if (current.equals(","))
		       {
			   advance();test();
			   if(!isStr()) throw new SyntaxException("add: string needed");
			   defs.add(extract());
			   advance();test();
		       }
		else defs.add(null);
		accept(")");
		advance();
		return true;
	    }
	else return false;
    }
    private final boolean namedefP() throws SyntaxException
    {
	if (current.equals("name"))
	    {
		advance();test();
		accept("=");
		advance();test();
		if(!isStr()) throw new SyntaxException("name: string needed");
		schema.name=extract();
		advance();
		return true;
	    }
	else return false;
    }
    private final boolean builtinP() throws SyntaxException
    {
	if (current.equals("builtin"))
	    {
		advance();test();
		accept("=");
		advance();test();
		if(!isStr()) throw new SyntaxException("builtin: string needed");
		try {
		    schema.builtin=Integer.parseInt(extract());
		} catch(NumberFormatException nfe) {
		    throw new SyntaxException("builtin: NaN: "+nfe.getMessage());
		}
		advance();
		return true;
	    }
	else return false;
    }
    private final boolean uridefP() throws SyntaxException
    {
	if (current.equals("URI"))
	    {
		advance();test();
		accept("=");
		advance();test();
		if(!isStr()) throw new SyntaxException("URI: string needed");
		schema.URI=extract();
		advance();
		return true;
	    }
	else return false;
    }
}

/**
 *A really simple "tokenizer".  It helps somewhat.  Maybe.
 */
class SimpleToken
{
    char[] specials;//slow if big
    char[] whites;//slow if big
    char stringDelim;
    BufferedReader in;
    static final char bufsize = 32;
    char[] cbuf = new char[bufsize];
    int index;
    char current;
    boolean end;
    public SimpleToken(char[] whites, char[]specials, char stringDelim, BufferedReader in)
    {
	this.whites=whites;
	this.specials=specials;
	this.stringDelim=stringDelim;
	this.in=in;
	index=bufsize;
	end = !read();
    }
    private boolean isSpecial()
    {
	for(int i=0;i<specials.length;i++) if (current==specials[i]) return true;
	return false;
    }
    private boolean isWhite()
    {
	for(int i=0;i<whites.length;i++) if (current==whites[i]) return true;
	return false;
    }
    private boolean read()
    {
	if(index==bufsize) if (!reload()){ end=true; return false;}
	current=cbuf[index++];
	return true;
    }
    private boolean reload()
    {//warning, discards cbuf!
	int temp;
	try{
	    temp=in.read(cbuf,0,bufsize);
	} catch(IOException e){temp=-1;}
	  
	if(temp==-1) return false;
	if(temp==bufsize) {index=0;return true;}
	index=bufsize-temp;
	for(int i=1;i<=temp;i++) cbuf[bufsize-i]=cbuf[temp-i];
	return true;
    }
    public String nextToken()
    {
	StringBuffer retval=new StringBuffer();
	if(end) return null;
	do {
	    if(!isWhite()) break;
	} while(read());
	if(end) return null;
	if (current==stringDelim)
	    {
		retval.append(current);
		read();  //if fails, another '"' is appended and then '""' is returned... OK!
		do {
		    retval.append(current);
		    if(current==stringDelim) break;
		} while(read());
		read();//get rid of closing stringDelim
		return new String(retval);
	    }
	if (isSpecial())
	    {
		retval.append(current);
		read();
		return new String(retval);
	    }
	do{
	    if(current==stringDelim || isWhite() || isSpecial()) break;
	    retval.append(current);
	} while(read());
	return new String(retval);
    }
}
